"""
ZSE API Module

User-facing interfaces:
- cli: Command-line interface (zse command)
- server: FastAPI REST server (OpenAI compatible)
- webui: Web-based playground and admin dashboard
"""

__all__ = [
    "cli",
    "server",
    "webui",
]
